@@ -20,6 +20,7 @@
 #include <linux/ncp_fs.h>
 #include <asm/segment.h>
 #include <linux/errno.h>
+#include <linux/locks.h>
 #include "ncplib_kernel.h"
 
 struct ncp_dirent {
@@ -200,12 +201,30 @@ static int                c_size;
 static int                c_seen_eof;
 static int                c_last_returned_index;
 static struct ncp_dirent* c_entry = NULL;
+static int                c_lock = 0;
+static struct wait_queue *c_wait = NULL;
+
+static inline void
+ncp_lock_dircache(void)
+{
+	while (c_lock)
+		sleep_on(&c_wait);
+	c_lock = 1;
+}
+
+static inline void
+ncp_unlock_dircache(void)
+{
+	c_lock = 0;
+	wake_up(&c_wait);
+}
 
 static int
 ncp_readdir(struct inode *inode, struct file *filp,
 	    struct dirent *dirent, int count)
 {
-	int result, i = 0;
+	int result = 0;
+	int i = 0;
         int index = 0;
 	struct ncp_dirent *entry = NULL;
         struct ncp_server *server = NCP_SERVER(inode);
@@ -238,6 +257,7 @@ ncp_readdir(struct inode *inode, struct file *filp,
 		return -EIO;
 	}
 
+	ncp_lock_dircache();
 	if (c_entry == NULL) 
 	{
 		i = sizeof (struct ncp_dirent) * NCP_READDIR_CACHE_SIZE;
@@ -245,7 +265,8 @@ ncp_readdir(struct inode *inode, struct file *filp,
 		if (c_entry == NULL)
 		{
 			printk("ncp_readdir: no MEMORY for cache\n");
-			return -ENOMEM;
+			result = -ENOMEM;
+			goto finished;
 		}
 	}
 
@@ -255,10 +276,11 @@ ncp_readdir(struct inode *inode, struct file *filp,
 		if (filldir(dirent,".",1, filp->f_pos,
 			    ncp_info_ino(server, dir)) < 0)
 		{
-			return 0;
+			goto finished;
 		}
 		filp->f_pos += 1;
-		return ROUND_UP(NAME_OFFSET(dirent)+i+1);
+		result = ROUND_UP(NAME_OFFSET(dirent)+i+1);
+		goto finished;
         }
 
 	if (filp->f_pos == 1)
@@ -266,10 +288,11 @@ ncp_readdir(struct inode *inode, struct file *filp,
 		if (filldir(dirent,"..",2, filp->f_pos,
 			    ncp_info_ino(server, dir->dir)) < 0)
 		{
-			return 0;
+			goto finished;
 		}
 		filp->f_pos += 1;
-		return ROUND_UP(NAME_OFFSET(dirent)+i+1);
+		result = ROUND_UP(NAME_OFFSET(dirent)+i+1);
+		goto finished;
 	}
 
 	if ((inode->i_dev == c_dev) && (inode->i_ino == c_ino))
@@ -286,42 +309,44 @@ ncp_readdir(struct inode *inode, struct file *filp,
 		}
                 if ((entry == NULL) && c_seen_eof)
 		{
-                        return 0;
+			goto finished;
 		}
 	}
 
 	if (entry == NULL)
 	{
+		int entries;
 		DDPRINTK("ncp_readdir: Not found in cache.\n");
 
 		if (ncp_is_server_root(inode))
 		{
-			result = ncp_read_volume_list(server, filp->f_pos,
-						      NCP_READDIR_CACHE_SIZE);
-			DPRINTK("ncp_read_volume_list returned %d\n", result);
+			entries = ncp_read_volume_list(server, filp->f_pos,
+						       NCP_READDIR_CACHE_SIZE);
+			DPRINTK("ncp_read_volume_list returned %d\n", entries);
 
 		}
 		else
 		{
-			result = ncp_do_readdir(server, inode, filp->f_pos,
-						NCP_READDIR_CACHE_SIZE,
-						c_entry);
-			DPRINTK("ncp_readdir returned %d\n", result);
+			entries = ncp_do_readdir(server, inode, filp->f_pos,
+						 NCP_READDIR_CACHE_SIZE,
+						 c_entry);
+			DPRINTK("ncp_readdir returned %d\n", entries);
 		}
 
-		if (result < 0)
+		if (entries < 0)
 		{
 			c_dev = 0;
 			c_ino = 0;
-			return result;
+			result = entries;
+			goto finished;
 		}
 
-		if (result > 0)
+		if (entries > 0)
 		{
-                        c_seen_eof = (result < NCP_READDIR_CACHE_SIZE);
+                        c_seen_eof = (entries < NCP_READDIR_CACHE_SIZE);
 			c_dev  = inode->i_dev;
 			c_ino  = inode->i_ino;
-			c_size = result;
+			c_size = entries;
 			entry = c_entry;
                         c_last_returned_index = 0;
                         index = 0;
@@ -336,7 +361,7 @@ ncp_readdir(struct inode *inode, struct file *filp,
         if (entry == NULL)
 	{
                 /* Nothing found, even from a ncp call */
-                return 0;
+		goto finished;
         }
 
         if (index < c_size)
@@ -374,15 +399,18 @@ ncp_readdir(struct inode *inode, struct file *filp,
                 if (filldir(dirent, entry->i.entryName, entry->i.nameLen,
                             entry->f_pos, ino) < 0)
 		{
-                        return 0;
+			goto finished;
                 }
 
                 filp->f_pos += 1;
                 index += 1;
                 entry += 1;
-		return ROUND_UP(NAME_OFFSET(dirent)+i+1);
+		result = ROUND_UP(NAME_OFFSET(dirent)+i+1);
+		goto finished;
 	}
-	return 0;
+ finished:
+	ncp_unlock_dircache();
+	return result;
 }
 
 static int
@@ -800,7 +828,7 @@ ncp_lookup(struct inode *dir, const char *__name, int len,
 
 	memcpy(name, __name, len);
 	name[len] = 0;
-
+	lock_super(dir->i_sb);
         result_info = ncp_find_dir_inode(dir, name);
 
         if (result_info != 0)
@@ -814,6 +842,7 @@ ncp_lookup(struct inode *dir, const char *__name, int len,
                    inode number */
 
                 *result = iget(dir->i_sb, ncp_info_ino(server, result_info));
+		unlock_super(dir->i_sb);
                 iput(dir);
 
                 if (*result == NULL)
@@ -828,7 +857,9 @@ ncp_lookup(struct inode *dir, const char *__name, int len,
            server. */
 
         found_in_cache = 0;
-        
+
+        ncp_lock_dircache();
+
         if ((dir->i_dev == c_dev) && (dir->i_ino == c_ino))
 	{
                 int first = c_last_returned_index;
@@ -851,6 +882,7 @@ ncp_lookup(struct inode *dir, const char *__name, int len,
                 }
 		while (i != first);
         }
+	ncp_unlock_dircache();
 
         if (found_in_cache == 0)
 	{
@@ -873,6 +905,7 @@ ncp_lookup(struct inode *dir, const char *__name, int len,
 		}
 		if (res != 0)
 		{
+			unlock_super(dir->i_sb);
                         iput(dir);
                         return -ENOENT;
                 }
@@ -883,10 +916,12 @@ ncp_lookup(struct inode *dir, const char *__name, int len,
 
 	if (!(*result = ncp_iget(dir, &finfo)))
 	{
+		unlock_super(dir->i_sb);
 		iput(dir);
 		return -EACCES;
 	}
 
+	unlock_super(dir->i_sb);
 	iput(dir);
 	return 0;
 }
@@ -916,6 +951,7 @@ ncp_create(struct inode *dir, const char *name, int len, int mode,
 	_name[len] = '\0';
 	str_upper(_name);
 
+	lock_super(dir->i_sb);
 	if (ncp_open_create_file_or_subdir(NCP_SERVER(dir),
 					   NCP_ISTRUCT(dir), _name,
 			 		   OC_MODE_CREATE|OC_MODE_OPEN|
@@ -923,6 +959,7 @@ ncp_create(struct inode *dir, const char *name, int len, int mode,
 					   0, AR_READ|AR_WRITE,
 					   &finfo) != 0)
 	{
+		unlock_super(dir->i_sb);
 		iput(dir);
 		return -EACCES;
 	}
@@ -935,10 +972,12 @@ ncp_create(struct inode *dir, const char *name, int len, int mode,
 	if (!(*result = ncp_iget(dir, &finfo)) < 0)
 	{
 		ncp_close_file(NCP_SERVER(dir), finfo.file_handle);
+		unlock_super(dir->i_sb);
 		iput(dir);
 		return -EINVAL;
 	}
 
+	unlock_super(dir->i_sb);
 	iput(dir);
 	return 0;	
 }
@@ -955,6 +994,7 @@ ncp_mkdir(struct inode *dir, const char *name, int len, int mode)
 		|| (   (len == 2)
 		    && (name[1] == '.'))))
 	{
+		iput(dir);
 		return -EEXIST;
 	}
 
