@@ -237,6 +237,10 @@ ipx_sap_find_nearest(int server_type, struct sockaddr_ipx *result,
 
 	if ((sock = socket(AF_IPX,SOCK_DGRAM,PF_IPX)) < 0)
 	{
+		if (errno == EINVAL)
+		{
+			return NCPL_ET_NO_IPX;
+		}
 		return errno;
 	}
 	
@@ -255,6 +259,10 @@ ipx_sap_find_nearest(int server_type, struct sockaddr_ipx *result,
 	
 	if(bind(sock,(struct sockaddr*)&addr,sizeof(addr))==-1)
 	{
+		if (errno == EADDRNOTAVAIL)
+		{
+			errno = NCPL_ET_NO_INTERFACE;
+		}
 		goto finished;
 	}
 	
@@ -340,7 +348,11 @@ ipx_make_reachable(IPXNet network)
 
 	if (sock == -1)
 	{
-		return -1;
+		if (errno == EINVAL)
+		{
+			return NCPL_ET_NO_IPX;
+		}
+		return errno;
 	}
 
 	opt=1;
@@ -516,14 +528,16 @@ do_ncp_call(struct ncp_conn *conn, int request_size)
 		*((struct ncp_request_header *)(&(conn->packet)));
 
 	int result;
-	int retries = 3;
+	int retries = 20;
 	int len;
 	long err;
-	struct ncp_reply_header *r =
-		(struct ncp_reply_header *)&(conn->packet);
 
 	while (retries > 0)
 	{
+		struct ncp_reply_header reply;
+		struct sockaddr_ipx sender;
+		int sizeofaddr = sizeof(sender);
+
 		retries -= 1;
 
 		result = sendto(conn->ncp_sock, conn->packet,
@@ -537,30 +551,48 @@ do_ncp_call(struct ncp_conn *conn, int request_size)
 		}
 
 	re_select:
-		len = ipx_recv(conn->ncp_sock,
-			       conn->packet, NCP_PACKET_SIZE, 0, 1, &err);
+		len = ipx_recvfrom(conn->ncp_sock,
+				   (char *)&reply, sizeof(reply),
+				   MSG_PEEK, &sender, &sizeofaddr, 3, &err);
 
-		if (   (len == sizeof(*r))
-		    && (r->type == NCP_POSITIVE_ACK))
+		if ((len < 0) && (err == ETIMEDOUT))
 		{
-			goto re_select;
-		}
-		if (   (len          >= sizeof(*r))
-		    && (r->type      == NCP_REPLY)
-		    && (   (request.type    == NCP_ALLOC_SLOT_REQUEST)
-			|| (   (r->sequence  == request.sequence)
-			    && (r->conn_low  == request.conn_low)
-			    && (r->conn_high == request.conn_high))))
-		{
-			conn->reply_size = len;
-			break;
+			continue;
 		}
 		if (len < 0)
 		{
 			return err;
 		}
+
+		if (   /* Is the sender wrong? */
+		       (memcmp(&sender.sipx_node,
+			       &(conn->i.addr.sipx_node), 6) != 0)
+		    || (sender.sipx_port != conn->i.addr.sipx_port)
+
+		       /* Did the sender send a positive acknowledge? */
+		    || (   (len == sizeof(reply))
+			&& (reply.type == NCP_POSITIVE_ACK))
+
+		       /* Did we get a bogus answer? */
+		    || (   (len          <  sizeof(reply))
+			|| (reply.type   != NCP_REPLY)
+			|| (   (request.type     != NCP_ALLOC_SLOT_REQUEST)
+			    && (   (reply.sequence  != request.sequence)
+				|| (reply.conn_low  != request.conn_low)
+				|| (reply.conn_high != request.conn_high)))))
+		{
+			/* Then throw away the packet */
+			ipx_recv(conn->ncp_sock, (char *)&reply, sizeof(reply),
+				 0, 1, &err);
+			goto re_select;
+		}
+
+		ipx_recv(conn->ncp_sock, conn->packet, NCP_PACKET_SIZE,
+			 0, 1, &err);
+		conn->reply_size = len;
+		return 0;
 	}
-	return 0;
+	return ETIMEDOUT;
 }
 
 static int
@@ -578,8 +610,9 @@ ncp_mount_request(struct ncp_conn *conn, int function)
 
 	if (conn->has_subfunction != 0)
 	{
-		*(__u16 *)&(h->data[0]) = conn->current_size
-			- sizeof(struct ncp_request_header)- 2;
+		*(__u16 *)&(h->data[0])
+			= htons(conn->current_size
+				- sizeof(struct ncp_request_header) - 2);
 	}
 
 	request.function   = function;
@@ -616,8 +649,9 @@ ncp_temp_request(struct ncp_conn *conn, int function)
 
 	if (conn->has_subfunction != 0)
 	{
-		*(__u16 *)&(h->data[0]) = conn->current_size
-			- sizeof(struct ncp_request_header) - 2;
+		*(__u16 *)&(h->data[0])
+			= htons(conn->current_size
+				- sizeof(struct ncp_request_header) - 2);
 	}
 
 	h->type = NCP_REQUEST;
@@ -753,9 +787,10 @@ ncp_connect_any(struct ncp_conn *conn, int wdog_needed)
 	char name[NCP_BINDERY_NAME_LEN];
 	long result;
 
-	if (ipx_sap_find_nearest(IPX_SAP_FILE_SERVER, &addr, name) != 0)
+	if ((result = ipx_sap_find_nearest(IPX_SAP_FILE_SERVER,
+					   &addr, name)) != 0)
 	{
-		return -1;
+		return result;
 	}
 
 	if ((result = ncp_connect_addr(conn, &addr, wdog_needed)) != 0)
@@ -1310,7 +1345,7 @@ ncp_fopen_nwc(const char *user, const char *mode)
 
 struct ncp_conn_spec *
 ncp_find_conn_spec(const char *server, const char *user, const char *password,
-		   uid_t uid, long *err)
+		   int login_necessary, uid_t uid, long *err)
 {
 	static struct ncp_conn_spec spec;
 
@@ -1407,6 +1442,12 @@ ncp_find_conn_spec(const char *server, const char *user, const char *password,
 		}
 	}
 
+	if (login_necessary == 0)
+	{
+		memset(spec.user, 0, sizeof(spec.user));
+		memset(spec.password, 0, sizeof(spec.password));
+	}
+
 	if (strlen(spec.user) == 0)
 	{
 		return &spec;
@@ -1524,7 +1565,8 @@ ncp_initialize_as(int *argc, char **argv,
 		i += 1;
 	}
 
-	spec = ncp_find_conn_spec(server, user, password, getuid(), err);
+	spec = ncp_find_conn_spec(server, user, password, login_necessary,
+				  getuid(), err);
 
 	if (spec == NULL)
 	{
@@ -1821,7 +1863,7 @@ ncp_send_broadcast(struct ncp_conn *conn,
 }
 
 /*
- * result is a 8-byte buffer
+ * target is a 8-byte buffer
  */
 long
 ncp_get_encryption_key(struct ncp_conn *conn,
@@ -1878,6 +1920,27 @@ ncp_get_bindery_object_id(struct ncp_conn *conn,
 	return 0;
 }
 
+long
+ncp_get_bindery_object_name(struct ncp_conn *conn,
+			    __u32 object_id,
+			    struct ncp_bindery_object *target)
+{
+	long result;
+	ncp_init_request_s(conn, 54);
+	ncp_add_dword(conn, htonl(object_id));
+
+	if ((result = ncp_request(conn, 23)) != 0) {
+		ncp_unlock_conn(conn);
+		return result;
+	}
+
+	target->object_id   = ntohl(ncp_reply_dword(conn, 0));
+	target->object_type = ntohs(ncp_reply_word (conn, 4));
+	memcpy(target->object_name, ncp_reply_data(conn, 6), 48);
+	ncp_unlock_conn(conn);
+	return 0;
+}
+
 long
 ncp_scan_bindery_object(struct ncp_conn *conn,
 			__u32 last_id, __u16 object_type, char *search_string,
@@ -1906,6 +1969,40 @@ ncp_scan_bindery_object(struct ncp_conn *conn,
 	return 0;
 }
 	
+long
+ncp_create_bindery_object(struct ncp_conn *conn,
+			  __u16 object_type,
+			  const char *object_name,
+			  __u8 object_security,
+			  __u8 object_status)
+{
+	long result;
+	ncp_init_request_s(conn, 50);
+	ncp_add_byte(conn, object_status);
+	ncp_add_byte(conn, object_security);
+	ncp_add_word(conn, htons(object_type));
+	ncp_add_pstring(conn, object_name);
+
+	result = ncp_request(conn, 23);
+	ncp_unlock_conn(conn);
+	return result;
+}
+
+
+long
+ncp_delete_bindery_object(struct ncp_conn *conn,
+			  __u16 object_type,
+			  const char *object_name)
+{
+	long result;
+	ncp_init_request_s(conn, 51);
+	ncp_add_word(conn, htons(object_type));
+	ncp_add_pstring(conn, object_name);
+
+	result = ncp_request(conn, 23);
+	ncp_unlock_conn(conn);
+	return result;
+}
 
 long
 ncp_read_property_value(struct ncp_conn *conn,
@@ -1932,6 +2029,149 @@ ncp_read_property_value(struct ncp_conn *conn,
 	return 0;
 }
 
+
+long
+ncp_scan_property(struct ncp_conn *conn,
+		  __u16 object_type, const char *object_name,
+		  __u32 last_id, char *search_string,
+	  	  struct ncp_property_info *property_info)
+{
+	long result;
+	ncp_init_request_s(conn, 60);
+	ncp_add_word(conn, htons(object_type));
+	ncp_add_pstring(conn, object_name);
+	ncp_add_dword(conn, htonl(last_id));
+	ncp_add_pstring(conn, search_string);
+
+	if ((result = ncp_request(conn, 23)) != 0) {
+		ncp_unlock_conn(conn);
+		return result;
+	}
+
+	memcpy(property_info->property_name,ncp_reply_data(conn, 0), 16);
+	property_info->property_flags = ncp_reply_byte(conn,16);
+	property_info->property_security = ncp_reply_byte(conn,17);
+	property_info->search_instance = ntohl(ncp_reply_dword(conn,18));
+	property_info->value_available_flag = ncp_reply_byte(conn,22);
+	property_info->more_properties_flag = ncp_reply_byte(conn,23);
+	ncp_unlock_conn(conn);
+	return 0;
+}
+
+long
+ncp_add_object_to_set(struct ncp_conn *conn,
+		      __u16 object_type, const char *object_name,
+		      const char *property_name,
+		      __u16 member_type,
+		      const char *member_name)
+{
+	long result;
+	ncp_init_request_s(conn, 65);
+	ncp_add_word(conn, htons(object_type));
+	ncp_add_pstring(conn, object_name);
+	ncp_add_pstring(conn, property_name);
+	ncp_add_word(conn, htons(member_type));
+	ncp_add_pstring(conn, member_name);
+
+	result = ncp_request(conn, 23);
+	ncp_unlock_conn(conn);
+	return result;
+}
+
+long
+ncp_change_property_security(struct ncp_conn *conn,
+			     __u16 object_type, const char *object_name,
+		  	     const char *property_name,
+			     __u8 property_security)
+{
+	long result;
+	ncp_init_request_s(conn, 59);
+	ncp_add_word(conn, htons(object_type));
+	ncp_add_pstring(conn, object_name);
+	ncp_add_byte(conn, property_security);
+	ncp_add_pstring(conn, property_name);
+
+	result = ncp_request(conn, 23);
+	ncp_unlock_conn(conn);
+	return result;
+}
+
+long
+ncp_create_property(struct ncp_conn *conn,
+		     __u16 object_type, const char *object_name,
+	  	     const char *property_name,
+		     __u8 property_flags, __u8 property_security)
+{
+	long result;
+	ncp_init_request_s(conn, 57);
+	ncp_add_word(conn, htons(object_type));
+	ncp_add_pstring(conn, object_name);
+	ncp_add_byte(conn, property_flags);
+	ncp_add_byte(conn, property_security);
+	ncp_add_pstring(conn, property_name);
+
+	result = ncp_request(conn, 23);
+	ncp_unlock_conn(conn);
+	return result;
+}
+
+long
+ncp_delete_object_from_set(struct ncp_conn *conn,
+			   __u16 object_type, const char *object_name,
+			   const char *property_name,
+			   __u16 member_type,
+			   const char *member_name)
+{
+	long result;
+	ncp_init_request_s(conn, 66);
+	ncp_add_word(conn, htons(object_type));
+	ncp_add_pstring(conn, object_name);
+	ncp_add_pstring(conn, property_name);
+	ncp_add_word(conn, htons(member_type));
+	ncp_add_pstring(conn, member_name);
+
+	result = ncp_request(conn, 23);
+	ncp_unlock_conn(conn);
+	return result;
+}
+
+long 
+ncp_delete_property(struct ncp_conn *conn,
+		    __u16 object_type, const char *object_name,
+		    const char *property_name)
+{
+	long result;
+	ncp_init_request_s(conn, 58);
+	ncp_add_word(conn, htons(object_type));
+	ncp_add_pstring(conn, object_name);
+	ncp_add_pstring(conn, property_name);
+
+	result = ncp_request(conn, 23);
+	ncp_unlock_conn(conn);
+	return result;
+}
+
+long
+ncp_write_property_value(struct ncp_conn *conn,
+		         __u16 object_type, const char *object_name,
+		         const char *property_name,
+			 __u8 segment,
+			 struct nw_property *property_value)
+{
+	long result;
+	ncp_init_request_s(conn, 62);
+	ncp_add_word(conn, htons(object_type));
+	ncp_add_pstring(conn, object_name);
+	ncp_add_byte(conn,segment);
+	ncp_add_byte(conn, property_value->more_flag);
+	ncp_add_pstring(conn, property_name);
+	ncp_add_mem(conn, property_value->value, 128);
+
+	result = ncp_request(conn, 23);
+	ncp_unlock_conn(conn);
+	return result;
+}
+
 long
 ncp_login_encrypted(struct ncp_conn *conn,
 		    const struct ncp_bindery_object *object,
@@ -1956,6 +2196,43 @@ ncp_login_encrypted(struct ncp_conn *conn,
 	return result;
 }
 
+long
+ncp_change_login_passwd(struct ncp_conn *conn,
+			const struct ncp_bindery_object *object,
+			const unsigned char *key,
+			const unsigned char *oldpasswd,
+			const unsigned char *newpasswd)
+{
+	long id = htonl(object->object_id);
+	unsigned char cryptkey[8];
+	unsigned char newpwd[16]; /* new passwd as stored by server */
+	unsigned char oldpwd[16]; /* old passwd as stored by server */
+	unsigned char len;
+	long result;
+
+	memcpy(cryptkey, key, 8);
+	shuffle((byte *)&id, oldpasswd, strlen(oldpasswd), oldpwd);
+	shuffle((byte *)&id, newpasswd, strlen(newpasswd), newpwd);
+	nw_encrypt(cryptkey, oldpwd, cryptkey);
+	newpassencrypt(oldpwd, newpwd, newpwd);
+	newpassencrypt(oldpwd+8, newpwd+8, newpwd+8);
+	if ((len = strlen(newpasswd)) > 63)
+	{
+		len = 63;
+	}
+	len = ((len ^ oldpwd[0] ^ oldpwd[1]) & 0x7f) | 0x40;
+
+	ncp_init_request_s(conn, 75);
+	ncp_add_mem(conn, cryptkey, 8);
+	ncp_add_word(conn, htons(object->object_type));
+	ncp_add_pstring(conn, object->object_name);
+	ncp_add_byte(conn, len);
+	ncp_add_mem(conn, newpwd, 16);
+	result = ncp_request(conn, 23);
+	ncp_unlock_conn(conn);
+	return result;
+}
+
 long
 ncp_login_user(struct ncp_conn *conn,
 	       const unsigned char *username,
@@ -2338,7 +2615,42 @@ ncp_delete_directory(struct ncp_conn *conn,
 	ncp_unlock_conn(conn);
 	return result;
 }
-	
+
+long
+ncp_add_trustee(struct ncp_conn *conn,
+		int dir_handle, const char *path,
+		__u32 object_id, __u8 rights)
+{
+	long result;
+
+	ncp_init_request_s(conn, 13);
+	ncp_add_byte(conn, dir_handle);
+	ncp_add_dword(conn, htonl(object_id));
+	ncp_add_byte(conn, rights);
+	ncp_add_pstring(conn, path);
+
+	result = ncp_request(conn, 22);
+	ncp_unlock_conn(conn);
+	return result;
+}
+
+long
+ncp_delete_trustee(struct ncp_conn *conn,
+		   int dir_handle, const char *path,__u32 object_id)
+{
+	long result;
+
+	ncp_init_request_s(conn, 14);
+	ncp_add_byte(conn, dir_handle);
+	ncp_add_dword(conn, htonl(object_id));
+	ncp_add_byte(conn, 0);
+	ncp_add_pstring(conn, path);
+
+	result = ncp_request(conn, 22);
+	ncp_unlock_conn(conn);
+	return result;
+}
+
 long
 ncp_rename_directory(struct ncp_conn *conn,
 		     int dir_handle,
@@ -2984,3 +3296,33 @@ ncp_alloc_short_dir_handle(struct ncp_conn *conn,
 	ncp_unlock_conn(conn);
 	return result;
 }
+
+long
+ncp_add_trustee_set(struct ncp_conn *conn,
+		    __u8 volume_number, __u32 dir_entry,
+		    __u16 rights_mask,
+		    int object_count, struct ncp_trustee_struct *rights)
+{
+	long result = 0;
+
+	ncp_init_request(conn);
+	ncp_add_byte(conn, 10); /* subfunction */
+	ncp_add_byte(conn, 0);  /* dos name space */
+	ncp_add_byte(conn, 0);	/* reserved */
+	ncp_add_word(conn, 0x8006);
+	ncp_add_word(conn, rights_mask);
+	ncp_add_word(conn, object_count);
+	ncp_add_handle_path(conn, volume_number, dir_entry, 1, NULL);
+
+	while (object_count > 0)
+	{
+		ncp_add_dword(conn, htonl(rights->object_id));
+		ncp_add_word(conn, rights->rights);
+		object_count -= 1;
+		rights += 1;
+	}
+
+	result = ncp_request(conn, 87);
+	ncp_unlock_conn(conn);
+	return result;
+}
