@@ -25,7 +25,7 @@ static void help(void);
 void
 main(int argc, char *argv[])
 {
-	struct ncp_conn conn;
+	struct ncp_conn *conn;
 
 	char default_queue[] = "*";
 	char *queue = default_queue;
@@ -42,10 +42,11 @@ main(int argc, char *argv[])
 
 	char *file_name;
 	int file;
+	long err;
 
 	progname = argv[0];
 
-	memzero(j); memzero(pj); memzero(q); memzero(conn);
+	memzero(j); memzero(pj); memzero(q);
 
 	if (   (argc == 2)
 	    && (strcmp(argv[1], "-h") == 0))
@@ -54,9 +55,9 @@ main(int argc, char *argv[])
 		exit(0);
 	}
 
-	if (ncp_initialize(&conn, &argc, argv, 1) != 0)
+	if ((conn = ncp_initialize(&argc, argv, 1, &err)) == NULL)
 	{
-		perror("Could not open connection");
+		com_err(argv[0], err, "in ncp_initialize");
 		exit(1);
 	}
 
@@ -82,7 +83,7 @@ main(int argc, char *argv[])
 		switch (opt) {
 		case 'h':
 			help();
-			ncp_close(&conn);
+			ncp_close(conn);
 			exit(1);
 		case 'p':
 			/* Path */
@@ -191,6 +192,7 @@ main(int argc, char *argv[])
 			{
 				printf("queue name too long: %s\n",
 				       optarg);
+				ncp_close(conn);
 				exit(1);
 			}
 			queue = optarg;
@@ -211,6 +213,7 @@ main(int argc, char *argv[])
 			
 		default:
 			usage();
+			ncp_close(conn);
 			exit(1);
 		}
 	}
@@ -218,6 +221,7 @@ main(int argc, char *argv[])
 	if (optind != argc-1)
 	{
 		usage();
+		ncp_close(conn);
 		exit(1);
 	}
 
@@ -233,6 +237,7 @@ main(int argc, char *argv[])
 		if (file < 0)
 		{
 			perror("could not open file");
+			ncp_close(conn);
 			exit(1);
 		}
 
@@ -264,17 +269,19 @@ main(int argc, char *argv[])
 
 	str_upper(queue);
 
-	if (ncp_scan_bindery_object(&conn, 0xffffffff, NCP_BINDERY_PQUEUE,
+	if (ncp_scan_bindery_object(conn, 0xffffffff, NCP_BINDERY_PQUEUE,
 				    queue, &q) != 0)
 	{
 		printf("could not find queue %s\n", queue);
-		return;
+		ncp_close(conn);
+		exit(1);
 	}
 	
-	if (ncp_create_queue_job_and_file(&conn, q.object_id, &j) != 0)
+	if (ncp_create_queue_job_and_file(conn, q.object_id, &j) != 0)
 	{
 		printf("create error\n");
-		return;
+		ncp_close(conn);
+		exit(1);
 	}
 
 	written = 0;
@@ -286,7 +293,7 @@ main(int argc, char *argv[])
 			break;
 		}
 
-		if (ncp_write(&conn, j.file_handle,
+		if (ncp_write(conn, j.file_handle,
 			      written, read_this_time, buf) < read_this_time)
 		{
 			break;
@@ -297,19 +304,18 @@ main(int argc, char *argv[])
 
 	close(file);
 
-	if (ncp_close_file_and_start_job(&conn, q.object_id, &j) != 0) {
+	if (ncp_close_file_and_start_job(conn, q.object_id, &j) != 0) {
 		printf("close error\n");
-		return;
-	}			
+	}
 
-	ncp_close(&conn);
+	ncp_close(conn);
+	return;
 }
 
 static void
 usage(void)
 {
 	fprintf(stderr, "usage: %s [options] file\n", progname);
-	exit(1);
 }
 
 static void
